package com.example.spaceaa;

import Apdu.CertificateRapdu;
import Apdu.ECPointRapdu;
import Apdu.EmptyRapdu;
import Apdu.MutualAuthRapdu;
import Apdu.NonceRapdu;
import Apdu.RAPDU;
import Exceptions.CertificateException;
import Exceptions.GetNonceException;
import Exceptions.KeyAggrementException;
import Exceptions.LastCommandException;
import Exceptions.MSESetATException;
import Exceptions.MapException;
import Exceptions.SelectException;
import Exceptions.TPcdException;
import Exceptions.TPiccException;
import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Arrays;
import org.spongycastle.math.ec.ECCurve;
import org.spongycastle.math.ec.ECPoint;

/**
 *
 * @author Wojciech Wodo
 */
public class SPACEAA {

    private ECCurve curve;
    private CardCommunication card;
    static final String curveName = "BrainpoolP256r1";
    private String testOid = "800A04007F00070202040202";
    private String taggedOid = "060A04007F00070202040202";
    private String spaceaaAID = "A000000001";
    private String domain = "84010D";
    private ECPoint sG;
    private ECPoint G;
    private SecureRandom rand;
    
    public SPACEAA(CardCommunication card){
        curve = org.spongycastle.asn1.teletrust.TeleTrusTNamedCurves.getByName(curveName).getCurve();
        G = org.spongycastle.asn1.teletrust.TeleTrusTNamedCurves.getByName(curveName).getG();
        this.card = card;
        rand = new SecureRandom();
    }
    
    
    public String start(String PIN) throws MSESetATException, GetNonceException, MapException, KeyAggrementException, TPcdException, TPiccException, LastCommandException, CertificateException, SelectException{
        byte[] rapduB;
        RAPDU rapdu;
        int dataLen;
        StringBuilder val = new StringBuilder();
        String result ="";
        
        //Select
        result += "Start SPACE|AA \n\n";
        result += "Selecting Applet\n";
        String sel = "00A40400"+Util.intToHexLen(spaceaaAID.length()/2)+spaceaaAID;
        result += "Send to Card: "+sel+"\n";
        rapduB = card.send(Util.stringHexToByteHex(sel));
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new EmptyRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new SelectException();
        }
        result += "\n";
        
        //MSE Set AT
        result+= "MSE:SET AT\n";
        String mse = "0022C1A4";
        String pinS = "830103";
        dataLen = (testOid.length()+domain.length()+pinS.length())/2;
        mse = mse + Util.intToHexLen(dataLen) + testOid + pinS + domain;
        result += "Send to Card: "+mse+"\n";
        rapduB = card.send(Util.stringHexToByteHex(mse));
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new EmptyRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new MSESetATException();
        }
        result+="\n";
        
        //Get Nonce
        result += "Get Nonce\n";
        String getNonce = "10860000027C0000";
        rapduB = card.send(Util.stringHexToByteHex(getNonce));
        result += "Send to Card: "+getNonce+"\n";
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new NonceRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new GetNonceException();
        }
        byte[] key_pi = Util.KDF_PI(PIN);
        byte[] sB = Util.dec(((NonceRapdu)rapdu).getNonce(), key_pi);
        BigInteger s = new BigInteger(1,sB);
        sG = G.multiply(s);
        result += "\n";
        
        
        //MAP Nonce
        result += "Mapping\n";
        BigInteger EPH_SK_PCD = new BigInteger(curve.getFieldSize(),rand);
        ECPoint EPH_PK_PCD = G.multiply(EPH_SK_PCD);
        String mapNonce = "10860000"+Util.intToHexLen(EPH_PK_PCD.getEncoded().length+4)+"7C"+Util.intToHexLen(EPH_PK_PCD.getEncoded().length+2)+"83"+
                Util.intToHexLen(EPH_PK_PCD.getEncoded().length)+Util.byteHextoStringHex(EPH_PK_PCD.getEncoded()).replace(" ", "");
        rapduB = card.send(Util.stringHexToByteHex(mapNonce));
        result += "Send to Card: "+mapNonce+"\n";
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new ECPointRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new MapException();
        }
        ECPoint EPH_PK_PICC = curve.decodePoint(((ECPointRapdu)rapdu).getPoint());
        ECPoint H = EPH_PK_PICC.multiply(EPH_SK_PCD);
        ECPoint GDash = H.add(sG);
        result += "\n";
        
        //Key Agreement
        result += "Key Agreement Phase\n";
        BigInteger SK_PCD = new BigInteger(curve.getFieldSize(),rand);
        ECPoint PK_PCD = GDash.multiply(SK_PCD);
        String keyAg = "10860000"+Util.intToHexLen(PK_PCD.getEncoded().length+4)+"7C"+Util.intToHexLen(PK_PCD.getEncoded().length+2)+"83"+
                Util.intToHexLen(PK_PCD.getEncoded().length)+Util.byteHextoStringHex(PK_PCD.getEncoded()).replace(" ", "");
        rapduB = card.send(Util.stringHexToByteHex(keyAg));
        result += "Send to Card: "+keyAg+"\n";
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new ECPointRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new KeyAggrementException();
        }
        ECPoint PK_PICC = curve.decodePoint(((ECPointRapdu)rapdu).getPoint());
        ECPoint KXY = PK_PICC.multiply(SK_PCD);
        String x = KXY.getX().toBigInteger().toString(16);
        x = "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"+x;
        x = x.substring(x.length()-(curve.getFieldSize()/8*2), x.length());
        byte[] K = Util.stringHexToByteHex(x);
        result += "\n";
        
        //Command Mutual Authentication
        result += "Mutual Authentication\n";
        String dataToTPcd = "7F49";
        String dataToTPicc = "7F49";
        val = new StringBuilder();
        val.append(taggedOid);
        val.append("86");
        val.append(Util.encodeLenToString(PK_PICC.getEncoded().length));
        val.append(Util.byteHextoStringHex(PK_PICC.getEncoded()).replace(" ",""));
        dataToTPcd = dataToTPcd + Util.encodeLenToString(val.toString().length()/2)+val.toString();
        byte[] TPcd = Util.mac(Util.stringHexToByteHex(dataToTPcd), Util.KDF_MAC(K));
        String TPcdCommand = "00860000"+Util.intToHexLen(TPcd.length+4)+"7C"+Util.intToHexLen(TPcd.length+2)+"85"+Util.intToHexLen(TPcd.length)+
                Util.byteHextoStringHex(TPcd).replace(" ", "");
        rapduB = card.send(Util.stringHexToByteHex(TPcdCommand));
        result += "Send to Card: "+TPcdCommand+"\n";
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new MutualAuthRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new TPcdException();
        }
        val = new StringBuilder();
        val.append(taggedOid);
        val.append("86");
        val.append(Util.encodeLenToString(PK_PCD.getEncoded().length));
        val.append(Util.byteHextoStringHex(PK_PCD.getEncoded()).replace(" ",""));
        dataToTPicc = dataToTPicc + Util.encodeLenToString(val.toString().length()/2)+val.toString();
        byte[] TPicc = Util.mac(Util.stringHexToByteHex(dataToTPicc), Util.KDF_MAC(K));
        byte[] CardTPicc = ((MutualAuthRapdu)rapdu).getTPicc();
        if(!Arrays.equals(TPicc, CardTPicc)){
            throw new TPiccException();
        }
        result+="\n";
        
        //General Authenticate
        result+="General Authenticate\n";
        String lastCommand = "10860000027C00";
        rapduB = card.send(Util.stringHexToByteHex(lastCommand));
        result += "Send to Card: "+lastCommand+"\n";
        result += "Received from Card: "+Util.byteHextoStringHex(rapduB).replace(" ", "")+"\n";
        rapdu = new CertificateRapdu(rapduB);
        if(!(rapdu.getSw1()==(byte)0x90 && rapdu.getSw2()==(byte)0x00)){
            throw new LastCommandException();
        }
        byte[] wB = Util.dec(((CertificateRapdu)rapdu).getW(), Util.KDF_ENC(K));
        byte[] cert = Util.dec(((CertificateRapdu)rapdu).getCertificate(), Util.KDF_ENC(K));
        if(wB.length!=curve.getFieldSize()/8){
            byte[] tmp = new byte[curve.getFieldSize()/8];
            for(int i=0;i<tmp.length;i++){
                tmp[i] = wB[i];
            }
            wB = tmp;
        }
        BigInteger w = new BigInteger(1,wB);
        ECPoint Public_Key_Card = EPH_PK_PICC.multiply(w);
        
        if(!Util.byteHextoStringHex(cert).startsWith(Util.byteHextoStringHex(Public_Key_Card.getEncoded()))){
            throw new CertificateException();
        }
        return result;
    }
}
